#!/usr/bin/env python3

import base64
import os

import cachetools
import requests

from emojifs import __repository__, __version__


def set_user_agent(x):
    """Sets a reasonable User-Agent. x should be a Request headers-like object."""
    x['User-Agent'] = f"emojifs/{__version__} ({__repository__}) {x['User-Agent']}"


def getuid():
    """Windows doesn't have os.getuid; WinFsp is happy with -1 though."""
    return os.getuid() if hasattr(os, 'getuid') else -1


def getgid():
    """Windows doesn't have os.getgid; WinFsp is happy with -1 though."""
    return os.getgid() if hasattr(os, 'getgid') else -1


@cachetools.cached(cachetools.LRUCache(maxsize=200))
def get_emoji_bytes(url: str) -> bytes:
    """Returns the bytes for a given emoji URL.  Handles HTTP(S) and data URLs."""
    if url.startswith('http'):
        r = _session.get(url)
        r.raise_for_status()
        return r.content
    elif url.startswith('data:'):
        (prefix, data) = url.split(',', maxsplit=1)
        if not prefix.endswith('base64'):
            raise ValueError
        return base64.b64decode(data)


@cachetools.cached(cachetools.LRUCache(maxsize=20000))
def get_content_length(url: str) -> int:
    """Returns the size of an emoji.  Handles HTTP(S) and data URLs."""
    if url.startswith('http'):
        r = _session.head(url)
        r.raise_for_status()
        # Slack emojis are served over CloudFront which provides Content-Length.
        return int(r.headers['Content-Length'])
    elif url.startswith('data:'):
        (prefix, data) = url.split(',', maxsplit=1)
        if not prefix.endswith('base64'):
            raise ValueError
        padding = data[-2:].count('=')
        return int(3*len(data)/4 - padding)


_session = requests.Session()
set_user_agent(_session.headers)



#!/usr/bin/env python3
"""Stuff for doing stuff with Discord."""

__copyright__ = """
Copyright © 2020 Chris Danis

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import base64
import errno
import io
import operator
import stat
import time

import cachetools
import refuse.high as fuse
import requests
from logzero import logger

import emojifs.constants as constants
import emojifs.utils as utils


class Discord(fuse.LoggingMixIn, fuse.Operations):
    """A FUSE filesystem implementation for a Discord user's guilds' emojis."""

    _base_url = 'https://discord.com/api/v6/'

    def __init__(self, token: str):
        self._retry_after = {}  # URL -> time.time() after which it's ok to retry
        self._write_buffers = {}  # path (not name!) -> BytesIO

        # emoji metadata for 100 guilds + 1 guild membership list
        self._emojis_cache = cachetools.TTLCache(maxsize=100, ttl=600)
        self._membership_cache = cachetools.TTLCache(maxsize=1, ttl=600)

        self._session = requests.Session()
        utils.set_user_agent(self._session.headers)
        self._session.headers['Authorization'] = token

        r = self._request('GET', 'users/@me')
        j = r.json()
        logger.info('👍 Successfully authenticated to Discord as %s#%s', j['username'],
                    j['discriminator'])

    def _url(self, method):
        return f"{self._base_url}{method}"

    def _request(self, http_method, urlfrag, **kwargs):
        """Execute a request against our _session, respecting ratelimiting and raising if
        the response isn't okay.  Retry on ratelimiting (after sleeping) but not on other error."""
        # TODO: This doesn't actually line up with Discord's ratelimiting semantics all that well.
        #       Discord also returns an X-RateLimit-Bucket header -- a bucket can span across many
        #       possible URLs -- but there's not an obvious way to map from a URL to a bucket name
        #       without first being ratelimited for it (and then remembering that??)
        # TODO: Even if we were strictly following the vague recommendations, it wouldn't be
        #       sufficient: Discord special-cases emoji methods with per-guild limits, and doesn't
        #       expose that in any useful way in the API (which is weird, because you'd think that
        #       each guild could just be a different bucket?)
        #       See also https://discord.com/developers/docs/topics/rate-limits
        # TODO: this is very close to, but not quite, Slack._request().
        url = self._url(urlfrag)

        # Attempt to respect any known ratelimiting on the given URL path
        time.sleep(max(0, self._retry_after.get(url, 0) - time.time()))
        resp = self._session.request(http_method, url, **kwargs)
        if resp.status_code == 429:
            self._retry_after[url] = time.time() + resp.headers.get('X-RateLimit-Reset-After', 60)
            logger.warn('Got ratelimited by Discord; retrying after %s seconds for %s',
                        self._retry_after[url], url)
            return self._request(http_method, url, **kwargs)
        else:
            resp.raise_for_status()
        if resp.status_code == 204:
            logger.debug('resp for %s to %s: HTTP %s', http_method, url, resp.status_code)
        else:
            logger.debug('resp for %s to %s json: %s', http_method, url, resp.json())
        return resp

    @cachetools.cachedmethod(operator.attrgetter('_membership_cache'))
    def _get_guilds(self):
        '''Returns all the guilds we're a member of.'''
        j = self._request('GET', 'users/@me/guilds').json()
        return {g['id']: g for g in j}

    # NB: If you change the signature of this function, you must also update _invalidate_guild!
    @cachetools.cachedmethod(operator.attrgetter('_emojis_cache'))
    def _get_emojis(self, id: str):
        '''Returns all the emoji for a given guild.'''
        j = self._request('GET', f'guilds/{id}/emojis').json()
        return {e['name']: e for e in j}

    def _invalidate_guild(self, id: str):
        '''Clear the cache of a given guild.'''
        k = cachetools.keys.hashkey(id)
        if k in self._emojis_cache:
            del self._emojis_cache[k]

    def _emoji_url(self, e):
        extension = 'gif' if e['animated'] else 'png'
        return f"https://cdn.discordapp.com/emojis/{e['id']}.{extension}"

    def _emoji_filename(self, e) -> str:
        extension = 'gif' if e['animated'] else 'png'
        return f"{e['name']}.{extension}"

    # TODO: support aliases for guilds, with the alias becoming the 'primary' render if present

    # TODO: a guild-ignorelist and a guild-onlylist?

    def _guild_to_path(self, g) -> str:
        '''Render a guild object as a pathname component.'''
        # TODO: alias support
        if isinstance(g, dict):
            return self._guild_to_path(g['name'])
        if isinstance(g, str):
            return g.replace('/', '_')
        raise ValueError

    def _path_to_guild(self, path):
        '''Given a /discord/foo/bar path, find the guild matching foo.'''
        gh = path.split('/', maxsplit=2)[1]
        guilds = self._get_guilds()
        # Direct ID lookups
        if gh in guilds:
            return guilds[gh]
        # TODO: alias support
        # Match by names
        candidates = [g for g in guilds.values() if gh == self._guild_to_path(g)]
        if candidates:
            return candidates[0]
        return None

    def _path_to_emojiname(self, path):
        '''/discord/foo/bar.png --> bar'''
        split = path.split('/', maxsplit=2)
        if len(split) < 3 or not split[2]:
            return None
        # Strip off any file extension.
        eh = split[2].split('.', maxsplit=1)[0]
        return eh

    def _path_to_guildmoji(self, path):
        '''Given a /discord/foo/bar.png path, find and return (guild, emoji) objects.'''
        g = self._path_to_guild(path)
        if g:
            eh = self._path_to_emojiname(path)
            if eh:
                emojis = self._get_emojis(g['id'])
                if eh in emojis:
                    return (g, emojis[eh])
                else:
                    raise fuse.FuseOSError(errno.ENOENT)
            else:
                return (g, None)
            # TODO: Better document (& test!) the semantics of this.
        else:
            return (None, None)

    def _guild_is_writable(self, g) -> bool:
        MANAGE_EMOJIS = 0x40000000
        return bool(g['permissions'] & MANAGE_EMOJIS)

    def getattr(self, path, fh):
        if path == '/':
            return dict(
                st_mode=stat.S_IFDIR | 0o555 | stat.S_IWUSR,
                st_mtime=time.time(),
                st_ctime=time.time(),
                st_atime=time.time(),
                st_nlink=2,
                st_uid=utils.getuid(),
                st_gid=utils.getgid(),
            )

        if path in self._write_buffers:
            return dict(
                st_mode=stat.S_IFREG | 0o600,
                st_atime=time.time(),
                st_ctime=time.time(),
                st_mtime=time.time(),
                st_nlink=1,
                st_uid=utils.getuid(),
                st_gid=utils.getgid(),
                st_size=len(self._write_buffers[path].getbuffer())
            )

        # This will raise a Fuse ENOENT if a nonexistent emoji name was specified in the path
        (g, e) = self._path_to_guildmoji(path)
        if g is None:
            raise fuse.FuseOSError(errno.ENOENT)
        elif e is None:
            return dict(
                st_mode=stat.S_IFDIR | 0o555 | (stat.S_IWUSR if self._guild_is_writable(g)
                                                else 0),
                st_mtime=time.time(),
                st_ctime=time.time(),
                st_atime=time.time(),
                st_nlink=2,
                st_uid=utils.getuid(),
                st_gid=utils.getgid(),
            )
        else:
            return dict(
                st_mode=stat.S_IFREG | 0o444,
                st_atime=time.time(),
                st_ctime=time.time(),
                st_mtime=time.time(),
                st_nlink=1,
                st_uid=utils.getuid(),
                st_gid=utils.getgid(),
                st_size=utils.get_content_length(self._emoji_url(e))
            )

    def readdir(self, path, fh=None):
        rv = ['.', '..']
        if path == '/':
            rv.extend(self._guild_to_path(g) for g in self._get_guilds().values())
        (g, e) = self._path_to_guildmoji(path)
        if g is not None and e is None:
            rv.extend(self._emoji_filename(e) for e in self._get_emojis(g['id']).values())
        return rv

    def listxattr(self, path):
        if path == '/' or path in self._write_buffers:
            return []
        (g, e) = self._path_to_guildmoji(path)
        if not g and e:
            raise fuse.FuseOSError(errno.ENOENT)
        return [constants.URL_XATTR_NAME, constants.CREATEDBY_XATTR_NAME]

    def getxattr(self, path, attrname):
        if path == '/' or path in self._write_buffers:
            raise fuse.FuseOSError(errno.ENODATA)
        (g, e) = self._path_to_guildmoji(path)
        if not g and e:
            raise fuse.FuseOSError(errno.ENOENT)
        if attrname == constants.URL_XATTR_NAME:
            return bytes(self._emoji_url(e), 'utf-8')
        elif attrname == constants.CREATEDBY_XATTR_NAME:
            u = e['user']
            return bytes(f"{u['username']}#{u['discriminator']}", 'utf-8')
        else:
            raise fuse.FuseOSError(errno.ENODATA)

    def read(self, path, size, offset, fh):
        if path in self._write_buffers:
            b = self._write_buffers[path]
            b.seek(offset)
            return b.read(size)

        (g, e) = self._path_to_guildmoji(path)
        if g and e:
            b = utils.get_emoji_bytes(self._emoji_url(e))
            return b[offset:offset+size]
        else:
            raise fuse.FuseOSError(errno.ENOENT)

    def unlink(self, path):
        # TODO: what about an unlink on a new file open in _write_buffers ?
        (g, e) = self._path_to_guildmoji(path)
        if g and e:
            logger.info('🗑️ Deleting :%s: (id %s) from "%s" (id %s)',
                        e['name'], e['id'], g['name'], g['id'])
            self._request('DELETE', f"guilds/{g['id']}/emojis/{e['id']}")
            self._invalidate_guild(g['id'])
        elif g and not e:
            # Sorry, but we won't delete a whole discord.
            raise fuse.FuseOSError(errno.EPERM)
        else:
            raise fuse.FuseOSError(errno.ENOENT)

    def _path_to_extension(self, path):
        split = path.rsplit('.', maxsplit=1)
        if len(split) < 2:
            raise ValueError
        extension = split[-1]
        if extension.lower() not in ['jpg', 'jpeg', 'gif', 'png']:
            raise ValueError
        return extension

    def create(self, path, mode, fi=None):
        g = self._path_to_guild(path)
        if not g:
            raise fuse.FuseOSError(errno.ENOENT)
        if not self._guild_is_writable(g):
            raise fuse.FuseOSError(errno.EPERM)
        try:
            self._path_to_extension(path)
        except ValueError:
            raise fuse.FuseOSError(errno.EINVAL)
        self._write_buffers[path] = io.BytesIO()
        return 0

    def write(self, path, data, offset, fh):
        b = self._write_buffers[path]
        b.seek(offset)
        return b.write(data)

    def release(self, path, fh):
        if path in self._write_buffers:
            b = self._write_buffers[path]
            b.seek(0)
            g = self._path_to_guild(path)
            eh = self._path_to_emojiname(path)
            ext = self._path_to_extension(path)
            try:
                image_mime = 'jpeg' if ext == 'jpg' else ext
                payload = dict(
                    name=eh,
                    image=(f"data:image/{image_mime};base64,"
                           + base64.b64encode(b.getvalue()).decode('ascii'))
                )
                logger.info('📸 Creating :%s: on "%s" (id %s)', eh, g['name'], g['id'])
                self._request('POST', f"guilds/{g['id']}/emojis", json=payload)
            finally:
                b.close()
                del self._write_buffers[path]
                self._invalidate_guild(g['id'])

    # TODO: it looks like HTTP PATCH is supported, maybe we can re-write emojis in place??



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['emojifs']

package_data = \
{'': ['*']}

install_requires = \
['cachetools>=4,<5',
 'logzero>=1.5.0,<2.0.0',
 'refuse>=0,<1',
 'requests>=2,<3',
 'tomlkit>=0.6.0,<0.7.0']

entry_points = \
{'console_scripts': ['emojifs = emojifs.__main__:main']}

setup_kwargs = {
    'name': 'emojifs',
    'version': '0.2.0',
    'description': 'A FUSE filesystem to manipulate custom emojis on your various Slacks and Discords',
    'long_description': '<img align="right"\n  src="extras/ablobcouncil.gif"\n  title="yes! but no! but also, yes."\n  alt="the emojifs mascot: a blobmoji holding a \'yes\' sign as well as a \'no\' sign and emphatically waving both simultaneously">\n\n# emojifs\n\n## Introduction\n`emojifs` is a FUSE filesystem that allows you to manipulate custom emojis on your various Slacks and Discords.\n\n\n## Example usage applications\n\n    # Copy one emote from Slack to Discord.\n    cp /emoji/slack/xooglers/docker-fire.gif /emoji/discord/unchaind/\n\n    # Upload a whole pile of emoji.\n    cp ~/emojipacks/parrots/* /emoji/slack/mynewslack/\n\n    # Grab all those blobmoji across all my Discords and put them on Slack.\n    cp /emoji/discord/*/*blob* /emoji/slack/myslack/\n\n<!-- TODO xattr? -->\n\n## ⚠️ WARNING! ⚠️\n**☢️ 😱 DO NOT USE THIS PROGRAM. 😱 ☢️**  \nThis program is not a program of honor.  \n\nNo highly esteemed function is executed here.  \n\nWhat is here is dangerous and repulsive to us.  \n\nThe danger [is still present](https://support.discord.com/hc/en-us/articles/115002192352), in your time, as it was in ours,  \nwithout even the implied warranty of MERCHANTABILITY  \nor FITNESS FOR A PARTICULAR PURPOSE.  \n\nThis program is best shunned and left unused (but it is free software,  \nand you are welcome to redistribute it under certain conditions).  \n**😱 ☢️ DO NOT USE THIS PROGRAM. ☢️ 😱**  \n\n## Getting your Slack and Discord secrets\n\n**⚠️ SERIOUS WARNING! I MEAN IT! ⚠️**  \nDon\'t share your cookies or tokens with anyone or anything, including this program.  \nThey allow full control of your account!  \n**⚠️ SERIOUS WARNING! I MEAN IT! ⚠️**  \n\n### Slack\nFor Slack, the easiest thing to do is:\n* In your web browser, log in to all the Slacks you want to use with this hunk of junk.\n* Open your browser devtools network panel\n* Find a request going towards edgeapi.slack.com or api.slack.com or a URL path starting with /api.  In Chrome, you can just type `api` in the Filter box above the timeline.\n* Extract the cookie header value from your request, and use it in the config file below.\n\n### Discord\n* In your web browser, go to https://discord.com/app\n* Open your browser devtools network panel\n* Find a request going towards `https://discord.com/api/v6`.  In Chrome, you can just type `api` in the Filter box above the timeline.\n* Extract the value of the `authorization:` header from one of your requests, and use it in the config file below.\n\n<img align="left"\n  src="https://cdn.discordapp.com/emojis/739162266284458045.gif"\n  title="monkaTOS"\n  alt="Pepe the Frog, sweating, his eyes replaced with a blinking marquee reading TOS">\n**⚠️ SERIOUS WARNING! I MEAN IT! ⚠️**  \nUsing this program with Discord violates their [Terms of Service](https://support.discord.com/hc/en-us/articles/115002192352).  \nIt is possible your account could be banned.  \n**⚠️ SERIOUS WARNING! I MEAN IT! ⚠️**  \n\n&nbsp;  \n\n## Requirements\n* Python 3.7+\n* A system that supports libfuse2\n  * Linux or MacOS (or probably most BSDs, but I haven\'t tried)\n  * Windows might work with [WinFsp](https://github.com/billziss-gh/winfsp), also untested\n* A profound amount of either fearlessness, or foolishness, or both\n\n## Installation\n\nLike most Python packages, just `pip install emojifs`, or you can `git clone` the repo and `python setup.py install`.\n\nYou probably want to do either of the above inside a venv.\n\nIf you have [`poetry`](https://python-poetry.org/) installed, then `poetry install` will work great.\n\n## Writing a config file\n\nThe format is [TOML](https://toml.io/).  It looks a lot like an INI file, if that means something to you.\n\nThe default location for the configuration file is `~/.emojifs.toml`\n\n### A barebones config file\n\n```toml\n[emojifs]\nmountpoint = \'~/emoji\'\n\n[slack]\ncookies = [\'d=wpwQ4182w08qxmE4YP0gvlMb2L...\']\n\n[discord]\ntoken = \'mfa.x91xxxxx......\'\nacknowledged = "I understand that using this program violates Discord\'s ToS"\n```\n\nThat\'s all you need.  All Slacks you logged into will be autodetected.\n\nOn Windows, your mountpoint should be a drive letter:\n```toml\n[emojifs]\nmountpoint = \'E:\'\n```\n\n### A comprehensive configuration\nThere are a few niceties available:\n```toml\n[emojifs]\nmountpoint = \'/emoji\'  # if you want to feel extra cool\nforeground = true\n\n[slack]\ntokens = [\n  \'xoxp-asdf...\',\n  \'xoxs-qwer...\',\n]\n\ncookies = [\n  \'d=dvfib...\',\n  \'d=ivu80...\',\n]\n\n[slack.renames]\nthisisaverylongname = \'short\'\n\n```\n\nThis will:\n* mount everything under `/emoji`\n* keep emojifs in the foreground as it runs (necessary if you want verbose logging output)\n* first read the auth tokens one by one, then scrape logins for the cookies listed\n* instead of mounting `thisisaverylongname.slack.com`\'s emojis under the usual path, they\'ll appear under `/emoji/slack/short`.\n\n\n## Invoking emojifs\n\n```\nusage: emojifs [-h] [-m MOUNTPOINT] [-f FOREGROUND] [-c CONFIG] [-v] [-V]\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -m MOUNTPOINT, --mountpoint MOUNTPOINT\n                        Where to mount emojifs. If present here, overrides\n                        mountpoint from config. (default: None)\n  -f FOREGROUND, --foreground FOREGROUND\n                        If set, stay in the foreground. (default: False)\n  -c CONFIG, --config CONFIG\n                        Path to your config file with secrets\n                        (default: ~/.emojifs.toml)\n  -v, --verbose         Verbosity (-v, -vv, etc). Higher verbosities will log\n                        all HTTP traffic (NB: at higher levels, this will log\n                        your auth secrets!) (default: 0)\n  -V, --version         show program\'s version number and exit\n\n```\n\n\n## A note on semantics\nEmojis are always \'rendered\' in the filesystem with extensions (`.png`, `.gif`, etc) attached; however, the filesystem will accept reads and writes to filenames without extensions (assuming, of course, the filenames are valid emoji names).\n\n## Known issues\n* The first time you `ls` a directory for a Slack or Discord, it will take a very long time: possibly dozens of seconds 😬 but will be much faster afterwards.  Sorry, there are hopefully some reasonable ways to fix this.  (If you\'re morbidly curious, look for `real_sizes` in `emojifs/slack.py`.)\n* While deletions and creations are supported, overwriting emojis in place is *not* yet supported.  As a workaround, you can use `cp --remove-destination` which, before writing new versions, will delete any existent emojis.\n* Some Slacks have emojis from long ago which are aliases to emojis that don\'t themselves appear in the emoji listing -- although there isn\'t actually data missing, as they have a `data:` URL.\n* The use of this program with Discord violates their Terms of Service.\n* The existence of this program is an unforgivable sin.\n\n## Future work\n* Aliases for Discord guild names, as they can be unwieldly from the CLI.\n* EaaFS: Integrating popular emojipacks as a filesystem.\n\n## Acknowledgements\n* The lovely [Requests library](https://requests.readthedocs.io/)\n* [FUSE and its predecessors](https://en.wikipedia.org/wiki/Filesystem_in_Userspace)\n* [SlackPirate by emtunc](https://github.com/emtunc/SlackPirate)\n* [slackmoji by ksindi](https://github.com/ksindi/slackmoji) and [emojipacks by lambtron](https://github.com/lambtron/emojipacks)\n* [Cult of the Party Parrot](https://cultofthepartyparrot.com/)\n* [blobs.gg](https://blobs.gg/)\n',
    'author': 'Chris Danis',
    'author_email': 'cdanis@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cdanis/emojifs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
